use rocket::Response;
use rocket::http::Status;

use rocket_contrib::json::Json;

use trust_dns_client::client::ClientHandle;
use trust_dns_client::op::ResponseCode;
use trust_dns_client::rr::{DNSClass, RecordType};

use crate::{DbConn, models::dns};
use crate::models::errors::{ErrorResponse, make_500};
use crate::models::users::{LocalUser, UserInfo, Zone, AddZoneMemberRequest, CreateZoneRequest};


#[get("/zones/<zone>/records")]
pub async fn get_zone_records(
    mut client: dns::DnsClient,
    conn: DbConn,
    user_info: Result<UserInfo, ErrorResponse>,
    zone: dns::AbsoluteName
) -> Result<Json<Vec<dns::Record>>, ErrorResponse> {

    let user_info = user_info?;
    let zone_name = zone.to_string();

    conn.run(move |c| {
        if user_info.is_admin() {
            Zone::get_by_name(c, &zone_name)
        } else {
            user_info.get_zone(c, &zone_name)
        }
    }).await?;

    let response = {
        let query = client.query(zone.clone(), DNSClass::IN, RecordType::AXFR);
        query.await.map_err(make_500)?
    };

    if response.response_code() != ResponseCode::NoError {
        println!("Querrying AXFR of zone {} failed with code {}", *zone, response.response_code());
        return ErrorResponse::new(
            Status::NotFound,
            format!("Zone {} could not be found", *zone)
        ).err()
    }

    let answers = response.answers();
    let mut records: Vec<_> = answers.to_vec().into_iter()
        .map(dns::Record::from)
        .filter(|record| !matches!(record.rdata, dns::RData::NULL { .. } | dns::RData::DNSSEC(_)))
        .collect();

    // AXFR response ends with SOA, we remove it so it is not doubled in the response.
    records.pop();

    Ok(Json(records))
}

#[get("/zones")]
pub async fn get_zones(
    conn: DbConn,
    user_info: Result<UserInfo, ErrorResponse>,
) -> Result<Json<Vec<Zone>>, ErrorResponse> {
    let user_info = user_info?;

    let zones = conn.run(move |c| {
        if user_info.is_admin() {
            Zone::get_all(c)
        } else {
            user_info.get_zones(c)
        }
    }).await?;

    Ok(Json(zones))
}

#[post("/zones", data = "<zone_request>")]
pub async fn create_zone(
    conn: DbConn,
    mut client: dns::DnsClient,
    user_info: Result<UserInfo, ErrorResponse>,
    zone_request: Json<CreateZoneRequest>,
) -> Result<Json<Zone>, ErrorResponse> {
    user_info?.check_admin()?;


    // Check if the zone exists in the DNS server
    let response = {
        let query = client.query(zone_request.name.clone(), DNSClass::IN, RecordType::SOA);
        query.await.map_err(make_500)?
    };

    if response.response_code() != ResponseCode::NoError {
        println!("Querrying SOA of zone {} failed with code {}", *zone_request.name, response.response_code());
        return ErrorResponse::new(
            Status::NotFound,
            format!("Zone {} could not be found", *zone_request.name)
        ).err()
    }

    let zone = conn.run(move |c| {
        Zone::create_zone(c, zone_request.into_inner())
    }).await?;

    Ok(Json(zone))
}


#[post("/zones/<zone>/members", data = "<zone_member_request>")]
pub async fn add_member_to_zone<'r>(
    conn: DbConn,
    zone: dns::AbsoluteName,
    user_info: Result<UserInfo, ErrorResponse>,
    zone_member_request: Json<AddZoneMemberRequest>
) -> Result<Response<'r>, ErrorResponse> {
    let user_info = user_info?;
    let zone_name = zone.to_utf8();

    conn.run(move |c| {
        let zone = if user_info.is_admin() {
            Zone::get_by_name(c, &zone_name)
        } else {
            user_info.get_zone(c, &zone_name)
        }?;

        let new_member = LocalUser::get_user_by_uuid(c, &zone_member_request.id)?;
        zone.add_member(&c, &new_member)
    }).await?;

    Response::build()
        .status(Status::Created) // TODO: change this?
        .ok()
}
